"""Zope PlugIns Framework"""
from Globals import HTMLFile, default__class_init__, Persistent
from OFS.Folder import Folder
from OFS.ObjectManager import ObjectManager; from OFS.SimpleItem import Item
from AccessControl.PermissionRole import PermissionRole
from App.ProductContext import ProductContext
import Acquisition, operator, string, Products

"""What are PlugIns, and what does this pattern do?

PROBLEM -- You are creating an object which you want to be end-user
configurable and extensible.  You need to manage one or more sets of
user-selected collaborator objects which extend your base object.  These
collaborator objects may be of multiple types, although within a given set,
they must all support the same collaboration interface.  You don't know in
advance what types these objects will be, since you want users to be able
to create arbitrary implementations of your collaboration interface.  You
also want to store, add, and manage these collaborators like regular Zope
sub-objects (so that acquistion, etc. still work), but don't want the
collaborators to show up on the Folder add list.  Finally, for clarity,
you'd like each set of collaborators to have their own management tab
(ala Contents), but not show up on your Contents tab (if you have one).

SOLUTION -- PlugInContainers are Folder-derived objects which can have
specialized collaborators, namely PlugIns.  PlugInContainers have management
tabs (implemented by PlugInGroups) for each collaboration interface
(__plugin_kind__).  PlugIn classes are registered by kind, and
PlugInContainer classes have PlugInGroups which determine what kind(s)
of PlugIns can be added on each management tab.

EXAMPLE APPLICATIONS -- LoginManager has PlugInGroups for LoginMethods and
UserSources.  If ZCatalog used a PlugInGroup for Indexes using a plugin
kind of "ZCatalog Index", users could create arbitrary kinds of index
objects without touching ZCatalog itself.  And so on.

LIMITATIONS -- Due to the way PlugIn classes register their permissions
and constructors, they must be implemented directly in Python.  It is
theoretically possible to create specialized constructors that would do
the same magic for ZClasses, but it would appear to require Fulton-like
Zen to do so.  As Jim would say, "waaaa!"
"""
class PlugInBase:
    """
PlugIns are objects which collaborate with their container in some way.  They
notify their container when they are added or removed.  PlugIns have a 'kind'
(__plugin_kind__) which is a string that identifies their collaboration role.
    """
        
    isAPlugIn = 1

    __plugin_kind__ = ""

    def __init__(self,id='',title=''):
        self.id = id
        self.title = title

    def _setup(self):
        pass    # easy subclass extension stub
        
    def manage_afterAdd(self, item, container):
        self._setup()
        container = self.aq_parent
        if hasattr(container,'_addPlugIn'):
            container._addPlugIn(self)

    def manage_beforeDelete(self, item, container):
        container = self.aq_parent
        if hasattr(container,'_removePlugIn'):
            container._removePlugIn(self)


class PlugIn(PlugInBase, Item, Acquisition.Implicit, Persistent):
    """PlugIn that is compatible w/Zope management framework"""
    pass








class PlugInContainer(PlugInBase, Folder):
    """
PlugInContainers are Folders which can have specialized collaborators known as
PlugIns, managed under management tabs provided by PlugInGroups.

This base class is useless in and of itself, as you must subclass it
to create a particular kind of PlugInContainer.  Your subclass must
define attributes for each PlugInGroup you want to have, plus a
__plugin_groups__ attribute which contains all of the PlugInGroups.  Example:

 class MyPIC(PlugInContainer):
     UserSources  = PlugInGroup('UserSources',
         ['User Source'],attr='userSourcesList'
     )
     LoginMethods = PlugInGroup('LoginMethods',
         ['Login Method'], attr='loginMethodsList'
     )
     __plugin_groups__ = (UserSources, LoginMethods)
     ...
"""

    manage_main = HTMLFile('www/main', globals())

    def _addPlugIn(self, plugin):

        """Register a plugin with its group"""

        plugin = getattr(plugin, 'aq_base', plugin)
        kind = plugin.__plugin_kind__

        for group in self.__plugin_groups__:
            if kind in group.kinds:
                group._addPlugIn(self, plugin)
                break        
        else:
            raise TypeError,("Unsupported plugin kind '%s'" % plugin.__plugin_kind__)

        self.manage_refreshPlugIns()



    def _removePlugIn(self, plugin):

        """Unregister plugin from any supporting group(s)"""

        plugin = getattr(plugin, 'aq_base', plugin)
        kind = plugin.__plugin_kind__

        for group in self.__plugin_groups__:
            if kind in group.kinds:
                group._removePlugIn(self, plugin)
        self.manage_refreshPlugIns()
        

    def _findGroupFor(self, plugin):

        """Return the group a plugin belongs to"""

        for group in self.__plugin_groups__:
            group = group.__of__(self)
            if plugin in group._objectValues():
                return group


    def _installPlugIn(self, plugin, message="", REQUEST=None):

        """Install a plugin (called by plugin constructors)"""

        self._setObject(plugin.id, plugin)

        if REQUEST is not None:
            group = self._findGroupFor(plugin)
            return group.manage_main(group, REQUEST, update_menu=1,
                manage_tabs_message=message, URL='%s/%s/manage_workspace' % (REQUEST.URL1,group.id))








    def manage_refreshPlugIns(self, REQUEST=None):
        """Update plugin registries"""
        for group in self.__plugin_groups__:
            group.__of__(self).manage_refreshPlugIns()
            
        if REQUEST is not None:
            return self.manage_main(self, REQUEST,
                manage_tabs_message="PlugIns refreshed.")

    manage_options_left = {'label':'Methods', 'action':'manage_main'},

    manage_options_right = tuple(Folder.manage_options[1:-1])

    def manage_options(self):

        """Construct manage_options (tabs) from left+plugin groups+right"""

        def as_option(group):
            return {'label':group.title, 'action': group.id+'/manage_workspace'}

        return self.manage_options_left + \
            tuple(map(as_option,self.__plugin_groups__))+ \
            self.manage_options_right


    def all_meta_types(self,exclude_plugins=0):
        mt = Folder.all_meta_types.im_func(self)
        if not exclude_plugins:
            mt=list(mt)
            for g in self.__plugin_groups__:
                mt.extend(g.all_meta_types())
        return mt

    def filtered_meta_types(self,user,exclude_plugins=0):
        mt = Folder.filtered_meta_types.im_func(self,user)
        if exclude_plugins:
            return filter(lambda x: not x.has_key('plugin_add_method'), mt)
        return mt



    def manage_beforeDelete(self,item,container):
        PlugInBase.manage_beforeDelete.im_func(self,item,container)
        Folder.manage_beforeDelete(self,item,container)

    def manage_afterAdd(self,item,container):
        Folder.manage_afterAdd(self,item,container)
        PlugInBase.manage_afterAdd.im_func(self,item,container)
        

    __ac_permissions__ = (
        ('View management screens', ('manage_main','manage_refreshPlugIns')
        ),
    )

default__class_init__(PlugInContainer)


PluggableContainer = PlugInContainer    # Backward compatibility























def genericAddForm(pluginClass, globalDict):

    """Return a customized HTMLFile of the generic plugin Add form"""
    
    className = pluginClass.__name__
    formName = 'manage_add%sForm' % className
    
    f = globalDict[formName] = HTMLFile(
        'genericAddForm', globals(),
        title_       = pluginClass.meta_type,
        constructor_ = ('manage_add%s'     % className),
        __name__     = formName
    )

    return f


_defaultConstructor = '''
def manage_add%(name)s(self, id, title='', REQUEST=None):
    """Add %(meta_type)s"""
    ob = %(name)s(id, title)
    return self.Destination()._installPlugIn(ob, "%(meta_type)s added", REQUEST)
'''


def simpleConstructor(pluginClass,globalDict):
    name=pluginClass.__name__; meta_type = pluginClass.meta_type
    exec _defaultConstructor % locals() in globalDict
    return globalDict['manage_add%s' % name]


def defaultConstructors(pluginClass,globalDict):
    return (
        genericAddForm(pluginClass,globalDict),
        simpleConstructor(pluginClass,globalDict)
    )





class PlugInGroup(Acquisition.Implicit):
    """
PlugInGroups manage PlugIns of specific Kind(s) on behalf of a PlugInContainer.
The PlugIns are actually stored in the PlugInContainer, but the PlugInGroup
helps manage them as a group.
    """

    def __init__(self, id, kinds, attr, title=''):
        """
Create a PlugIn group to manage a registry of PlugIns of the specified kinds.
Registry will be stored in the PlugInContainer attribute specified by attr.
If attr is not specified, registry will be stored in the PlugInContainer's
__plugins__ attribute as a dictionary mapping kinds to lists of plugins.
        """

        self.id,self.kinds,self.attr,self.title = id,tuple(kinds),attr,title


    def all_meta_types(self):

        """Return meta_types which can be added to this group"""

        global PlugInKinds
   
        l = []
        for k in self.kinds:
            l.extend(PlugInKinds.get(k,[]))

        return l

    # We need filtered meta types, but we don't want to be an ObjectManager
    filtered_meta_types = ObjectManager.filtered_meta_types.im_func     # Hack!

    def _PlugInsChanged(self, container, pluginslist):
        """This is just a hookpoint for subclasses"""
        pass





    def manage_refreshPlugIns(self, REQUEST=None):
        """Refresh plugin registry"""
        container = self.aq_inner.aq_parent
        self._PlugInsChanged(container, getattr(container, self.attr, []))
        if REQUEST is not None:
            return self.manage_main(self, REQUEST,
                manage_tabs_message="PlugIns refreshed.",
            	URL=REQUEST.URL1+'/manage_workspace')

    def _objectValues(self):
        container = self.aq_inner.aq_parent
        return getattr(container, self.attr, [])
        
    def objectValues(self):
        """Get objects contained within the group"""
        container = self.aq_inner.aq_parent
        return map (lambda x,c=container: x.__of__(c), self._objectValues())
        
    def objectItems(self):
        """get (id, object) tuples of items in the group"""
        return map(lambda x: (x.id, x), self.objectValues())
        
    manage_main = manage_workspace = \
        HTMLFile('www/PlugInGroup', globals())

    manage_above = manage_below = ''

    def _addPlugIn(self, container, plugin):
        """Register plug-in with group"""
        l = getattr(container, self.attr, None) or []
        if plugin not in l: l.append(plugin)
        setattr(container, self.attr, l)

    def _removePlugIn(self, container, plugin):
        """Unregister plug-in from group"""
        l = getattr(container, self.attr, None) or []
        l = filter(lambda x,p=plugin: x is not p, l)
        setattr(container, self.attr, l)



    def tabs_path_info(self, script, path):
        path = self.absolute_url(relative=1) + '/foo'
        return self.aq_inner.aq_parent.tabs_path_info(script, path)

    def manage_first(self, ids=[], REQUEST=None):
        """Re-sort plugins, moving selected ones first"""

        container = self.aq_inner.aq_parent

        l = getattr(container, self.attr, None) or []

        l = filter(lambda x,ids=ids: x.id in ids, l) + \
            filter(lambda x,ids=ids: x.id not in ids, l)

        setattr(container, self.attr, l)
        container.manage_refreshPlugIns()

        if REQUEST:
            return self.manage_main(self,REQUEST,update_menu=1,
            	URL=REQUEST.URL1+'/manage_workspace')
            
    def manage_cutObjects(self, ids, REQUEST=None):
        """Put a reference to the objects named in ids in the clip board"""
        cp = self.aq_inner.aq_parent.manage_cutObjects(ids,REQUEST)
        if REQUEST is not None:
            return self.manage_main(self, REQUEST, cb_dataValid=1,
            	URL=REQUEST.URL1+'/manage_workspace')
        return cp
    
    def manage_copyObjects(self, ids, REQUEST=None, RESPONSE=None):
        """Put a reference to the objects named in ids in the clip board"""
        cp = self.aq_inner.aq_parent.manage_copyObjects(ids,REQUEST)
        if REQUEST is not None:
            return self.manage_main(self, REQUEST, cb_dataValid=1,
            	URL=REQUEST.URL1+'/manage_workspace')
        return cp





    def manage_renameObject(self, id, new_id, REQUEST=None):
        """Rename a particular sub-object"""
    	self.aq_inner.aq_parent.manage_renameObject(id,new_id,REQUEST)
        if REQUEST is not None:
            return self.manage_main(self, REQUEST, update_menu=1,
            	URL=REQUEST.URL1+'/manage_workspace')


    def manage_delObjects(self, ids=[], REQUEST=None):
        """Delete subordinate objects"""
    	self.aq_inner.aq_parent.manage_delObjects(ids,REQUEST)        
        if REQUEST is not None:
            return self.manage_main(self, REQUEST, update_menu=1,
            	URL=REQUEST.URL1+'/manage_workspace')


    def _constructPlugIn(self,name,*args,**kwargs):

        """Create a plug-in - used when container wants defaults plugins"""

        # Use filtered meta types so users can't bypass create permissions
        for meta in self.filtered_meta_types(self.REQUEST.AUTHENTICATED_USER):
            if meta['name']==name:
                method = meta['plugin_add_method']
                product = self.aq_inner.aq_parent.manage_addProduct[meta['product']]
                return apply(getattr(product, method), (product,)+args, kwargs)


    # XXX should have different permission for manage_first, but what?

    __ac_permissions__ = (
        ('View management screens',
            ('manage_first','manage_main','manage_workspace',
                'manage_above', 'manage_below', 'manage_refreshPlugIns')
        ),
    )

default__class_init__(PlugInGroup)



# Global PlugIn class registry
# key: kind name; value: lists of meta-type dicts for classes of kind
PlugInKinds = {}


def _registerPlugInClass(self, instance_class=None, meta_type='',
                          permission=None, constructors=(),
                          icon=None, permissions=None, legacy=(),):
    """
    Register a plug-in class.  The class registered will not be included in the
    master Products.meta_types registry, but only in the PlugInKinds registry.
    To register a plug-in class, create an initialize() method in your product's
    __init__.py.  In it, call context.registerPlugInClass() with the parameters
    you would use for a context.registerClass() call.
    """

    result = self.registerClass(instance_class,meta_type,
        permission,constructors,icon,permissions,legacy)

    meta = Products.meta_types[-1]
    Products.meta_types = Products.meta_types[:-1]

    if len(constructors)>1:
        method = constructors[1]
        if type(method) is type(()):
            name,method = method
        else: name = method.__name__
        meta['plugin_add_method'] = name

    kind = instance_class.__plugin_kind__

    l = PlugInKinds.get(kind, [])
    l.append(meta)
    PlugInKinds[kind] = l

    return result    

# Patch ourselves in so we look like part of the Zope registration API
ProductContext.registerPlugInClass = _registerPlugInClass















   

