from OFS.PropertySheets import PropertySheet
from Providers import Provider
from Globals import HTMLFile, default__class_init__
from PlugIns import defaultConstructors

psKey = 'ZPatterns.SheetProviders', 'PropertySheets'



































class SheetProvider(Provider):
    """
    Sheet provider example class.  This sheet provider uses the 'slot'
    facility of Racks/Rackmountables to provide arbitrary propertysheets
    stored persistently.  All of its methods would need to be redefined
    for SheetProviders that implement fixed-schema and/or non-slot-based
    property sheets.
    """

    __plugin_kind__ = 'Property Sheet Provider'


    def namesForRegistration(self,container):
        """
Return a tuple (names,namespaces) where names is a sequence of
sheet names, and namespaces a sequence of XML namespaces, which this sheet
provider supports.  '*' may be included in either or both sequences, which
means the provider will be called if a more specific provider could not
be found for a given sheet.  This is mainly so that sheet providers can
potentially create propertysheets dynamically based on the requested name
or namespace.
        """
        return {
            'provides': ('sheets',),
            'sheet_names': self.Sheet_Names,
            'sheet_xmlns': self.Sheet_Namespaces
        }


    def _PropertySheetFor(self,client,name='',xmlns=''):
        """
        Return client's propertysheet matching name or xmlns, if available.
        Otherwise return None.
        """
        pslist = client._v_readableSlot.get(psKey,())
        for ps in pslist:
            if (name and ps.id==name) or (xmlns and ps.xml_namespace()==xmlns):
                return ps



    def _addPropertySheetFor(self,client,name='',xmlns=''):
        """
        Return client's new propertysheet matching name or xmlns, if possible.
        Otherwise return None.
        """
        ps=PropertySheet(name, {'xmlns':xmlns})
        pslist = client._v_readableSlot.get(psKey,())+(ps,)
        client._v_writeableSlot[psKey] = pslist
        return ps


    def _delPropertySheetFor(self,client,name='',xmlns=''):
        """
        Delete client's propertysheet matching name or xmlns, if present.
        Always returns None (unless an unexpected error occurs, of course)
        """
        pslist = client._v_readableSlot.get(psKey,())
        newlist = tuple(filter(
            lambda ps,name=name,xmlns=xmlns: 
                (not name or ps.id!=name) and 
                (not xmlns or ps.xml_namespace()!=xmlns),
            pslist
        ))
        if newlist!=pslist:
            client._v_writeableSlot[psKey] = newlist


    def _objectDeleting(self,client):
        """Delete all of client's propertysheets"""
        if client._v_readableSlot.has_key(psKey):
            del client._v_writeableSlot[psKey]


    def _PropertySheetsFor(self, client):
        """Return a (possibly empty) list of PropertySheet objects for client"""
        # Note: this could include duplicate sheets if persistent sheets
        # are defined by more than one provider!
        return client._v_readableSlot.get(psKey,())



    # Class Metadata
    
    meta_type='Persistent Sheet Provider'
    
    Sheet_Names = ('*',)
    Sheet_Namespaces = ('*',)

    _properties=(
        {'id':'title', 'type': 'string', 'mode': 'w'},
        {'id':'Sheet_Names', 'type': 'tokens', 'mode': 'w'},
        {'id':'Sheet_Namespaces', 'type': 'lines', 'mode': 'w'},
    )





























class AcquiredSheetProvider(SheetProvider):
    """Delegate providing of certain sheets to parent ProviderContainer"""
    
    def namesForRegistration(self,container):
        c=container.aq_parent
        r=getattr(c,'_getRegistry',None)

        if r is None:
            raise TypeError, "No providers available to acquire in /%s" % c.absolute_url(1)

        return {
            'provides': ('sheets',),
            
            'sheet_names':
                self._handledSheets(
                    self.Sheet_Names, r('sheet_names').keys()
                ),
                
            'sheet_xmlns':
                self._handledSheets(
                    self.Sheet_Namespaces, r('sheet_xmlns').keys()
                )
        }
        
    def _handledSheets(self, filter, names):
        if '*' in filter: return names
        if '*' in names:  return filter
        return filter(lambda x,f=filter: x in f, names)

    def _PropertySheetFor(self,client,name='',xmlns=''):
        return self.aq_parent.aq_parent._PropertySheetFor(client,name,xmlns)

    def _addPropertySheetFor(self,client,name='',xmlns=''):
        return self.aq_parent.aq_parent._addPropertySheetFor(client,name,xmlns)

    def _delPropertySheetFor(self,client,name='',xmlns=''):
        return self.aq_parent.aq_parent._delPropertySheetFor(client,name,xmlns)




    def _objectAdded(self,client):
        return self.aq_parent.aq_parent._objectAdding(client)

    def _objectDeleted(self,client):
        return self.aq_parent.aq_parent._objectDeleting(client)


    def _PropertySheetsFor(self, client):
        # Note: this could include duplicate sheets if persistent sheets
        # are defined by more than one provider!
        return self.aq_parent.aq_parent._PropertySheetsFor(client)

    meta_type = 'Acquired Sheet Provider'




























def initialize(context):

    context.registerPlugInClass(
        SheetProvider,
        permission = 'Add Sheet Providers',
        constructors = defaultConstructors(SheetProvider,globals()),
        icon = 'www/sheetprov.gif'
    )

    #context.registerPlugInClass(
    #    AcquiredSheetProvider,
    #    permission = 'Add Acquired Sheet Providers',
    #    constructors = defaultConstructors(AcquiredSheetProvider,globals()),
    #    icon = 'www/acqsheetprov.gif'
    #)










