class Transactional:
    """
    Mix-in Adapter to simplify ZODB Transaction messages

    Whenever your subclass does work which requires transactional behavior, 
    it should call self._register(), to ensure it is registered with the current
    transaction.  Your subclass will then be able to receive the following
    (translated) messages from the ZODB transaction:

    _checkpoint() -- 
        apply work done so far (this is where committing should mostly go)

    _revert() -- 
        abort work done since last checkpoint or since the transaction began
        if not yet checkpointed

    _rollback() -- 
        abort all checkpointed work (this call is always preceded by a _revert()
        call to handle any work which was not yet checkpointed, and is always
        followed by a _cleanup() call, so don't make it do the work of either.)

    _vote() -- 
        raise an exception here if you want to force the transaction to abort

    _finalize() -- 
        make all work done permanent (try not to refer to objects' states here)

    _cleanup() --
        called whenever a transaction is terminated, whatever its outcome

    Your revert, rollback, finalize, and cleanup methods should not raise any 
    exceptions.  An untrapped exception from _finalize() (or _cleanup() after
    _finalize()) will put the entire ZODB application into "hosed" state which
    can only be undone by an application restart.  The ZODB transaction will
    in most cases silently toss exceptions from your _revert() and _rollback()
    routines, so it would be a good idea to log them somewhere.
    """

    _v_registered = None      # Are we registered w/a transaction?


    def _register(self):
        if self._v_registered: return
        get_transaction().register(Reporter(self))
        self._v_registered = 1

    def _unregister(self):
        del self._v_registered


    # From here down, override as you see fit...
    
    def _checkpoint(self):
        """Called during commit() phase of transaction/subtransaction;
           should do all the "real work" of committing"""
        pass

    def _revert(self):
        """Do the real work of aborting anything here"""
        pass

    def _rollback(self):
        """Called during tpc_abort - should do as little as possible"""
        pass

    def _vote(self):
        """Opportunity to raise an error to prevent commit"""
        pass

    def _finalize(self):
        """Called during tpc_finish - should do as little as possible"""
        pass

    def _cleanup(self):
        """Called after transaction is effectively *over*, so don't
           do anything here that either 1) isn't idempotent, or 2)
           requires knowledge about the state/data of the transaction."""
        pass




class Reporter:
    # Event translator for Transactionals
    
    tpc_entered = 0        # Are we in tpc yet?

    def __init__(self, client):
        self.client = client

    def tpc_begin(self, transaction, subtransaction=None):
        self.tpc_entered = 1

    def tpc_abort(self,transaction):
        try:
            self.client._revert()
            self.client._rollback()
        finally:
            self.end_tran()

    def tpc_vote(self,transaction):
        self.client._vote()

    def tpc_finish(self,transaction):
        try:
            self.client._finalize()
        finally:
            self.end_tran()

    def commit_sub(self, transaction):
        # Note tpc entered, just in case
        self.tpc_begin(transaction)

    def abort_sub(self, transaction):
        self.tpc_abort()
        
    def commit(self, object, transaction):
        self.client._checkpoint()





    def abort(self, object, transaction):
        try:
            self.client._revert()
        finally:
            if not self.tpc_entered:
                # Abort before tpc_begin() means transaction aborted early
                self.end_tran()


    def end_tran(self):
        try:
            self.client._cleanup()
        finally:
            self.client._unregister()



























from UserDict import UserDict

class TransientMapping(Transactional, UserDict):
    """A mapping that clears itself after every transaction"""

    saved = None

    def __setitem__(self, key, item):
        self._register()
        UserDict.__setitem__.im_func(self, key, item)

    def __delitem__(self, key):
        self._register()
        UserDict.__delitem__.im_func(self, key)

    def update(self, dict):
        self._register()
        UserDict.update.im_func(self, dict)

    def clear(self):
        self._register()
        UserDict.clear.im_func(self)

    def copy(self):
        return self.data.copy()

    def _checkpoint(self):
        self.saved = self.data.copy()
        
    def _revert(self):
        if self.saved:
            self.data = self.saved
            del self.saved

    # No need to define rollback, since cleanup always follows

    def _cleanup(self):
        UserDict.clear.im_func(self)
        Transactional._cleanup.im_func(self)
